# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

from paasta_tools.adhoc_tools import AdhocJobConfig
from paasta_tools.adhoc_tools import load_adhoc_job_config
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.utils import DeploymentsJsonV2


TEST_SERVICE_NAME = "example_happyhour"
TEST_SOA_DIR = "fake_soa_dir"
TEST_CLUSTER_NAME = "cluster"


def create_test_service():
    return PaastaServiceConfigLoader(
        service=TEST_SERVICE_NAME, soa_dir=TEST_SOA_DIR, load_deployments=True
    )


def deployment_json():
    return DeploymentsJsonV2(
        service="test-service",
        config_dict={
            "deployments": {
                "cluster.non_canary": {
                    "docker_image": "some_image",
                    "git_sha": "some_sha",
                    "image_version": None,
                },
                "cluster.canary": {
                    "docker_image": "some_image",
                    "git_sha": "some_sha",
                    "image_version": None,
                },
            },
            "controls": {
                "example_happyhour:%s.sample_batch"
                % TEST_CLUSTER_NAME: {"desired_state": "start", "force_bounce": None},
                "example_happyhour:%s.interactive"
                % TEST_CLUSTER_NAME: {"desired_state": "start", "force_bounce": None},
                f"{TEST_SERVICE_NAME}:{TEST_CLUSTER_NAME}.main": {
                    "desired_state": "start",
                    "force_bounce": None,
                },
                f"{TEST_SERVICE_NAME}:{TEST_CLUSTER_NAME}.canary": {
                    "desired_state": "start",
                    "force_bounce": None,
                },
                f"{TEST_SERVICE_NAME}:{TEST_CLUSTER_NAME}.example_child_job": {
                    "desired_state": "start",
                    "force_bounce": None,
                },
                f"{TEST_SERVICE_NAME}:{TEST_CLUSTER_NAME}.sample_batch": {
                    "desired_state": "start",
                    "force_bounce": None,
                },
                f"{TEST_SERVICE_NAME}:{TEST_CLUSTER_NAME}.interactive": {
                    "desired_state": "start",
                    "force_bounce": None,
                },
            },
        },
    )


def kubernetes_cluster_config():
    """Return a sample dict to mock paasta_tools.utils.load_service_instance_configs"""
    return {
        "main": {
            "instances": 3,
            "deploy_group": "{cluster}.non_canary",
            "cpus": 0.1,
            "mem": 1000,
        },
        "canary": {
            "instances": 1,
            "deploy_group": "{cluster}.canary",
            "cpus": 0.1,
            "mem": 1000,
        },
        "not_deployed": {
            "instances": 1,
            "deploy_group": "not_deployed",
            "cpus": 0.1,
            "mem": 1000,
        },
    }


def adhoc_cluster_config():
    return {
        "sample_batch": {
            "deploy_group": "{cluster}.non_canary",
            "cpus": 0.1,
            "mem": 1000,
            "cmd": "/bin/sleep 5s",
        },
        "interactive": {"deploy_group": "{cluster}.non_canary", "mem": 1000},
        "not_deployed": {"deploy_group": "not_deployed"},
    }


@patch(
    "paasta_tools.paasta_service_config_loader.load_service_instance_configs",
    autospec=True,
)
def test_kubernetes_instances(mock_load_service_instance_configs):
    mock_load_service_instance_configs.return_value = kubernetes_cluster_config()
    s = create_test_service()
    assert list(s.instances(TEST_CLUSTER_NAME, KubernetesDeploymentConfig)) == [
        "main",
        "canary",
        "not_deployed",
    ]
    mock_load_service_instance_configs.assert_called_once_with(
        service=TEST_SERVICE_NAME,
        instance_type="kubernetes",
        cluster=TEST_CLUSTER_NAME,
        soa_dir=TEST_SOA_DIR,
    )


@patch(
    "paasta_tools.paasta_service_config_loader.load_v2_deployments_json", autospec=True
)
@patch(
    "paasta_tools.paasta_service_config_loader.load_service_instance_configs",
    autospec=True,
)
def test_kubernetes_instances_configs(
    mock_load_service_instance_configs, mock_load_deployments_json
):
    mock_load_service_instance_configs.return_value = kubernetes_cluster_config()
    mock_load_deployments_json.return_value = deployment_json()
    s = create_test_service()
    expected = [
        KubernetesDeploymentConfig(
            service=TEST_SERVICE_NAME,
            cluster=TEST_CLUSTER_NAME,
            instance="main",
            config_dict={
                "port": None,
                "monitoring": {},
                "deploy": {},
                "data": {},
                "smartstack": {},
                "dependencies": {},
                "instances": 3,
                "deploy_group": f"{TEST_CLUSTER_NAME}.non_canary",
                "cpus": 0.1,
                "mem": 1000,
            },
            branch_dict={
                "docker_image": "some_image",
                "desired_state": "start",
                "force_bounce": None,
                "git_sha": "some_sha",
                "image_version": None,
            },
            soa_dir=TEST_SOA_DIR,
        ),
        KubernetesDeploymentConfig(
            service=TEST_SERVICE_NAME,
            cluster=TEST_CLUSTER_NAME,
            instance="canary",
            config_dict={
                "port": None,
                "monitoring": {},
                "deploy": {},
                "data": {},
                "smartstack": {},
                "dependencies": {},
                "instances": 1,
                "deploy_group": f"{TEST_CLUSTER_NAME}.canary",
                "cpus": 0.1,
                "mem": 1000,
            },
            branch_dict={
                "docker_image": "some_image",
                "desired_state": "start",
                "force_bounce": None,
                "git_sha": "some_sha",
                "image_version": None,
            },
            soa_dir=TEST_SOA_DIR,
        ),
    ]
    assert (
        list(s.instance_configs(TEST_CLUSTER_NAME, KubernetesDeploymentConfig))
        == expected
    )
    mock_load_service_instance_configs.assert_called_once_with(
        service=TEST_SERVICE_NAME,
        instance_type="kubernetes",
        cluster=TEST_CLUSTER_NAME,
        soa_dir=TEST_SOA_DIR,
    )
    mock_load_deployments_json.assert_called_once_with(
        TEST_SERVICE_NAME, soa_dir=TEST_SOA_DIR
    )


@patch(
    "paasta_tools.paasta_service_config_loader.load_v2_deployments_json", autospec=True
)
@patch(
    "paasta_tools.paasta_service_config_loader.load_service_instance_configs",
    autospec=True,
)
def test_adhoc_instances_configs(
    mock_load_service_instance_configs, mock_load_deployments_json
):
    mock_load_service_instance_configs.return_value = adhoc_cluster_config()
    mock_load_deployments_json.return_value = deployment_json()
    s = create_test_service()
    expected = [
        AdhocJobConfig(
            service=TEST_SERVICE_NAME,
            cluster=TEST_CLUSTER_NAME,
            instance="sample_batch",
            config_dict={
                "port": None,
                "monitoring": {},
                "deploy": {},
                "data": {},
                "smartstack": {},
                "dependencies": {},
                "cmd": "/bin/sleep 5s",
                "deploy_group": "cluster.non_canary",
                "cpus": 0.1,
                "mem": 1000,
            },
            branch_dict={
                "docker_image": "some_image",
                "desired_state": "start",
                "force_bounce": None,
                "git_sha": "some_sha",
                "image_version": None,
            },
            soa_dir=TEST_SOA_DIR,
        ),
        AdhocJobConfig(
            service=TEST_SERVICE_NAME,
            cluster=TEST_CLUSTER_NAME,
            instance="interactive",
            config_dict={
                "port": None,
                "monitoring": {},
                "deploy": {},
                "data": {},
                "smartstack": {},
                "dependencies": {},
                "deploy_group": "cluster.non_canary",
                "mem": 1000,
            },
            branch_dict={
                "docker_image": "some_image",
                "desired_state": "start",
                "force_bounce": None,
                "git_sha": "some_sha",
                "image_version": None,
            },
            soa_dir=TEST_SOA_DIR,
        ),
    ]
    for i in s.instance_configs(TEST_CLUSTER_NAME, AdhocJobConfig):
        print(i, i.cluster)
    assert list(s.instance_configs(TEST_CLUSTER_NAME, AdhocJobConfig)) == expected
    mock_load_service_instance_configs.assert_called_once_with(
        service=TEST_SERVICE_NAME,
        instance_type="adhoc",
        cluster=TEST_CLUSTER_NAME,
        soa_dir=TEST_SOA_DIR,
    )
    mock_load_deployments_json.assert_called_once_with(
        TEST_SERVICE_NAME, soa_dir=TEST_SOA_DIR
    )


@patch(
    "paasta_tools.paasta_service_config_loader.load_v2_deployments_json", autospec=True
)
@patch("paasta_tools.adhoc_tools.load_v2_deployments_json", autospec=True)
@patch(
    "paasta_tools.paasta_service_config_loader.load_service_instance_configs",
    autospec=True,
)
@patch(
    "paasta_tools.adhoc_tools.load_service_instance_config",
    autospec=True,
)
def test_old_and_new_ways_load_the_same_adhoc_configs(
    mock_adhoc_tools_load_service_instance_config,
    mock_load_service_instance_configs,
    mock_adhoc_tools_load_deployments_json,
    mock_load_deployments_json,
):
    mock_load_service_instance_configs.return_value = adhoc_cluster_config()
    mock_adhoc_tools_load_service_instance_config.side_effect = [
        adhoc_cluster_config().get("sample_batch"),
        adhoc_cluster_config().get("interactive"),
    ]
    mock_load_deployments_json.return_value = deployment_json()
    mock_adhoc_tools_load_deployments_json.return_value = deployment_json()
    s = create_test_service()
    expected = [
        load_adhoc_job_config(
            service=TEST_SERVICE_NAME,
            instance="sample_batch",
            cluster=TEST_CLUSTER_NAME,
            load_deployments=True,
            soa_dir=TEST_SOA_DIR,
        ),
        load_adhoc_job_config(
            service=TEST_SERVICE_NAME,
            instance="interactive",
            cluster=TEST_CLUSTER_NAME,
            load_deployments=True,
            soa_dir=TEST_SOA_DIR,
        ),
    ]
    assert list(s.instance_configs(TEST_CLUSTER_NAME, AdhocJobConfig)) == expected



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.utils import _run
from paasta_tools.utils import load_system_paasta_config


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "security-check",
        description="Performs a security check consisting of a few tests.",
        help="Performs a security check",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help='Name of service for which you wish to check. Leading "services-", as included in a '
        "Jenkins job name, will be stripped.",
        required=True,
        # strip any potential trailing / for folks tab-completing directories
        type=lambda x: x.rstrip("/"),
    )
    list_parser.add_argument(
        "-c", "--commit", help="Git sha of the image to check", required=True
    )
    list_parser.set_defaults(command=perform_security_check)


def perform_security_check(args):
    """It runs a few security tests, checks the return code and prints output that should help in fixing failures.
    If you are at Yelp, please visit https://confluence.yelpcorp.com/display/PAASTA/PaaSTA+security-check+explained
    to learn more.
    :param args: service - the name of the service; commit - upstream git commit.
    :return: 0 if the security-check passed, non-zero if it failed.
    """
    security_check_command = load_system_paasta_config().get_security_check_command()
    if not security_check_command:
        print("Nothing to be executed during the security-check step")
        return 0

    command = f"{security_check_command} {args.service} {args.commit}"

    ret_code, output = _run(command, timeout=3600, stream=True)
    if ret_code != 0:
        print(
            "The security-check failed. Please visit y/security-check-runbook to learn how to fix it ("
            "including whitelisting safe versions of packages and docker images)."
        )

    return ret_code



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import PaastaColors
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.deployment_utils import load_v2_deployments_json
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_service_docker_registry
from paasta_tools.utils import list_services


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "get-docker-image",
        help="Gets the docker image URL for the deployment of a service",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help="Name of the service which you want to get the docker image for.",
        required=True,
        # strip any potential trailing / for folks tab-completing directories
        type=lambda x: x.rstrip("/"),
    ).completer = lazy_choices_completer(list_services)
    list_parser.add_argument(
        "-i",
        "-l",
        "--deploy-group",
        help='Name of the deploy group, like "prod".',
        required=True,
    ).completer = lazy_choices_completer(list_deploy_groups)
    list_parser.add_argument(
        "-d",
        "--soa-dir",
        help="A directory from which soa-configs should be read from",
        default=DEFAULT_SOA_DIR,
    )

    list_parser.set_defaults(command=paasta_get_docker_image)


def paasta_get_docker_image(args):
    service = args.service
    deploy_group = args.deploy_group
    soa_dir = args.soa_dir
    validate_service_name(service, soa_dir)

    deployments = load_v2_deployments_json(service=service, soa_dir=soa_dir)
    docker_image = deployments.get_docker_image_for_deploy_group(deploy_group)

    if not docker_image:
        print(
            PaastaColors.red(
                f"There is no {service} docker_image for {deploy_group}. Has it been deployed yet?"
            ),
            file=sys.stderr,
        )
        return 1
    else:
        registry_uri = get_service_docker_registry(service=service, soa_dir=soa_dir)
        docker_url = f"{registry_uri}/{docker_image}"
        print(docker_url)
        return 0



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import sys
import time

from pysensu_yelp import Status

from paasta_tools import monitoring_tools
from paasta_tools.cli.cmds.logs import scribe_env_to_locations
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config

try:
    from scribereader import scribereader
    from scribereader.clog.readers import StreamTailerSetupError
except ImportError:
    scribereader = None


OOM_EVENTS_STREAM = "tmp_paasta_oom_events"


def compose_check_name_for_service_instance(check_name, service, instance):
    return f"{check_name}.{service}.{instance}"


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=(
            "Check the %s stream and report to Sensu if"
            " there are any OOM events." % OOM_EVENTS_STREAM
        )
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-r",
        "--realert-every",
        dest="realert_every",
        type=int,
        default=1,
        help="Sensu 'realert_every' to use.",
    )
    parser.add_argument(
        "--check-interval",
        dest="check_interval",
        type=int,
        default=1,
        help="How often this check runs, in minutes.",
    )
    parser.add_argument(
        "--alert-threshold",
        dest="alert_threshold",
        type=int,
        default=1,
        help="Number of OOM kills required in the check interval to send an alert.",
    )
    parser.add_argument(
        "-s",
        "--superregion",
        dest="superregion",
        required=True,
        help="The superregion to read OOM events from.",
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Print Sensu alert events instead of sending them",
    )
    return parser.parse_args(args)


def read_oom_events_from_scribe(cluster, superregion, num_lines=1000):
    """Read the latest 'num_lines' lines from OOM_EVENTS_STREAM and iterate over them."""
    # paasta configs incls a map for cluster -> env that is expected by scribe
    log_reader_config = load_system_paasta_config().get_log_reader()
    cluster_map = log_reader_config["options"]["cluster_map"]
    scribe_env = cluster_map[cluster]

    # `scribe_env_to_locations` slightly mutates the scribe env based on whether
    # or not it is in dev or prod
    host, port = scribereader.get_tail_host_and_port(
        **scribe_env_to_locations(scribe_env),
    )
    stream = scribereader.get_stream_tailer(
        stream_name=OOM_EVENTS_STREAM,
        tailing_host=host,
        tailing_port=port,
        lines=num_lines,
        superregion=superregion,
    )
    try:
        for line in stream:
            try:
                j = json.loads(line)
                if j.get("cluster", "") == cluster:
                    yield j
            except json.decoder.JSONDecodeError:
                pass
    except StreamTailerSetupError as e:
        if "No data in stream" in str(e):
            pass
        else:
            raise e


def latest_oom_events(cluster, superregion, interval=60):
    """
    :returns: {(service, instance): [OOMEvent, OOMEvent,...] }
              if the number of events > 0
    """
    start_timestamp = int(time.time()) - interval
    res = {}
    for e in read_oom_events_from_scribe(cluster, superregion):
        if e["timestamp"] > start_timestamp:
            key = (e["service"], e["instance"])
            res.setdefault(key, set()).add(e.get("container_id", ""))
    return res


def compose_sensu_status(
    instance, oom_events, is_check_enabled, alert_threshold, check_interval
):
    """
    :param instance: InstanceConfig
    :param oom_events: a list of OOMEvents
    :param is_check_enabled: boolean to indicate whether the check enabled for the instance
    """
    interval_string = f"{check_interval} minute(s)"
    instance_name = f"{instance.service}.{instance.instance}"
    if not is_check_enabled:
        return (Status.OK, f"This check is disabled for {instance_name}.")
    if not oom_events:
        return (
            Status.OK,
            f"No oom events for {instance_name} in the last {interval_string}.",
        )
    elif len(oom_events) >= alert_threshold:
        return (
            Status.CRITICAL,
            f"The Out Of Memory killer killed processes for {instance_name} "
            f"in the last {interval_string}.",
        )
    else:
        # If the number of OOM kills isn't above the alert threshold,
        # don't send anything. This will keep an alert open if it's already open,
        # but won't start a new alert if there wasn't one yet
        return None


def send_sensu_event(instance, oom_events, args):
    """
    :param instance: InstanceConfig
    :param oom_events: a list of OOMEvents
    """
    check_name = compose_check_name_for_service_instance(
        "oom-killer", instance.service, instance.instance
    )
    monitoring_overrides = instance.get_monitoring()
    status = compose_sensu_status(
        instance=instance,
        oom_events=oom_events,
        is_check_enabled=monitoring_overrides.get("check_oom_events", True),
        alert_threshold=args.alert_threshold,
        check_interval=args.check_interval,
    )
    if not status:
        return

    memory_limit = instance.get_mem()
    try:
        memory_limit_str = f"{int(memory_limit)}MB"
    except ValueError:
        memory_limit_str = memory_limit

    monitoring_overrides.update(
        {
            "page": False,
            "alert_after": "0m",
            "realert_every": args.realert_every,
            "runbook": "y/check-oom-events",
            "tip": (
                "Follow the runbook to investigate and rightsize memory usage "
                f"(curr: {memory_limit_str})"
            ),
        }
    )
    return monitoring_tools.send_event(
        service=instance.service,
        check_name=check_name,
        overrides=monitoring_overrides,
        status=status[0],
        output=status[1],
        soa_dir=instance.soa_dir,
        dry_run=args.dry_run,
    )


def main(sys_argv):
    args = parse_args(sys_argv[1:])
    cluster = load_system_paasta_config().get_cluster()
    victims = latest_oom_events(
        cluster, args.superregion, interval=(60 * args.check_interval)
    )

    for (service, instance) in get_services_for_cluster(cluster, soa_dir=args.soa_dir):
        try:
            instance_config = get_instance_config(
                service=service,
                instance=instance,
                cluster=cluster,
                load_deployments=False,
                soa_dir=args.soa_dir,
            )
            oom_events = victims.get((service, instance), [])
            send_sensu_event(instance_config, oom_events, args)
        except NotImplementedError:  # When instance_type is not supported by get_instance_config
            pass


if __name__ == "__main__":
    main(sys.argv)



import logging
import os
import socket
import time
from abc import ABC
from abc import abstractmethod
from inspect import currentframe
from types import TracebackType
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

from typing_extensions import Protocol

from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)

try:
    import yelp_meteorite
except ImportError:
    yelp_meteorite = None

_metrics_interfaces: Dict[str, Type["BaseMetrics"]] = {}


class TimerProtocol(Protocol):
    def __enter__(self) -> "TimerProtocol":
        raise NotImplementedError()

    def __exit__(
        self,
        err_type: Optional[type],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        raise NotImplementedError()

    def __call__(
        self,
    ) -> Optional[float]:
        raise NotImplementedError()

    def start(self) -> None:
        raise NotImplementedError()

    def stop(self, **kwargs: Any) -> None:
        raise NotImplementedError()

    def record(self, value: float, **kwargs: Any) -> None:
        raise NotImplementedError()


class GaugeProtocol(Protocol):
    def set(self, value: Union[int, float]) -> None:
        raise NotImplementedError()


class CounterProtocol(Protocol):
    def count(self) -> None:
        raise NotImplementedError()


class BaseMetrics(ABC):
    def __init__(self, base_name: str) -> None:
        self.base_name = base_name

    @abstractmethod
    def create_timer(self, name: str, **kwargs: Any) -> TimerProtocol:
        raise NotImplementedError()

    @abstractmethod
    def create_gauge(self, name: str, **kwargs: Any) -> GaugeProtocol:
        raise NotImplementedError()

    @abstractmethod
    def create_counter(self, name: str, **kwargs: Any) -> CounterProtocol:
        raise NotImplementedError()

    @abstractmethod
    def emit_event(self, name: str, **kwargs: Any) -> bool:
        raise NotImplementedError()


def get_metrics_interface(base_name: str) -> BaseMetrics:
    metrics_provider = load_system_paasta_config().get_metrics_provider()
    return _metrics_interfaces[metrics_provider](base_name)


def register_metrics_interface(
    name: Optional[str],
) -> Callable[[Type[BaseMetrics]], Type[BaseMetrics]]:
    def outer(func: Type[BaseMetrics]) -> Type[BaseMetrics]:
        _metrics_interfaces[name] = func
        return func

    return outer


@register_metrics_interface("meteorite")
class MeteoriteMetrics(BaseMetrics):
    def __init__(self, base_name: str) -> None:
        self.base_name = base_name
        if yelp_meteorite is None:
            raise ImportError(
                "yelp_meteorite not imported, please try another metrics provider"
            )

    def create_timer(self, name: str, **kwargs: Any) -> TimerProtocol:
        # yelp_meteorite returns an EmptyMetric object if the timer is misconfigured
        # ...but that doesn't have the same interface ;_;
        return cast(
            Timer, yelp_meteorite.create_timer(self.base_name + "." + name, **kwargs)
        )

    def create_gauge(self, name: str, **kwargs: Any) -> GaugeProtocol:
        # yelp_meteorite returns an EmptyMetric object if the gauge is misconfigured
        # ...but that doesn't have the same interface ;_;
        return cast(
            Gauge, yelp_meteorite.create_gauge(self.base_name + "." + name, **kwargs)
        )

    def create_counter(self, name: str, **kwargs: Any) -> CounterProtocol:
        return yelp_meteorite.create_counter(self.base_name + "." + name, **kwargs)

    def emit_event(self, name: str, **kwargs: Any) -> bool:
        return yelp_meteorite.emit_event(self.base_name + "." + name, **kwargs)


class Timer(TimerProtocol):
    def __init__(self, name: str) -> None:
        self.name = name
        self.elapsed_ms: Optional[float] = None

    def __enter__(self) -> TimerProtocol:
        self.start()
        return self

    def __exit__(
        self,
        err_type: Optional[type],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if not err_type:
            self.stop()

    def __call__(self) -> Optional[float]:
        if self.elapsed_ms is None:
            self.stop()
        return self.elapsed_ms

    def start(self) -> None:
        log.debug("timer {} start at {}".format(self.name, time.time()))

    def stop(self, **kwargs: Any) -> None:
        log.debug("timer {} stop at {}".format(self.name, time.time()))

    def record(self, value: float, **kwargs: Any) -> None:
        log.debug(f"timer {self.name} record value {value}")


class Gauge(GaugeProtocol):
    def __init__(self, name: str) -> None:
        self.name = name

    def set(self, value: Union[int, float]) -> None:
        log.debug(f"gauge {self.name} set to {value}")


class Counter(CounterProtocol):
    def __init__(self, name: str) -> None:
        self.name = name
        self.counter = 0

    def count(self) -> None:
        self.counter += 1
        log.debug(f"counter {self.name} incremented to {self.counter}")


@register_metrics_interface(None)
class NoMetrics(BaseMetrics):
    def __init__(self, base_name: str) -> None:
        self.base_name = base_name

    def create_timer(self, name: str, **kwargs: Any) -> Timer:
        return Timer(self.base_name + "." + name)

    def create_gauge(self, name: str, **kwargs: Any) -> Gauge:
        return Gauge(self.base_name + "." + name)

    def create_counter(self, name: str, **kwargs: Any) -> Counter:
        return Counter(self.base_name + "." + name)

    def emit_event(self, name: str, **kwargs: Any) -> bool:
        log.debug(f"event {name} occurred with properties: {kwargs}")
        return True


def system_timer(
    name: str = "system_process_duration", dimensions: Dict[str, Any] = {}
) -> TimerProtocol:
    metrics_interface = get_metrics_interface("paasta")
    parent_pid = os.getppid()
    pid = os.getpid()
    hostname = socket.gethostname()
    # set our parent caller as the path dimension, e.g. /path/to/setup_kubernetes_job
    current = currentframe()
    parent = current.f_back
    path = parent.f_globals.get("__file__", "unknown")
    # Note any of these dimensions may be overridden by caller
    default_dimensions = {
        "path": path,
        "host": hostname,
        # ppid is included given some system processes are called multiple times in parallel for a single 'run'
        "parent_pid": parent_pid,
        "pid": pid,
    }
    default_dimensions.update(dimensions)
    return metrics_interface.create_timer(
        name=name, default_dimensions=default_dimensions
    )



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from typing import Collection
from typing import Dict
from typing import Generator
from typing import Mapping
from typing import Tuple

from humanize import naturaltime

from paasta_tools.cli.cmds.mark_for_deployment import can_user_deploy_service
from paasta_tools.cli.cmds.mark_for_deployment import get_deploy_info
from paasta_tools.cli.cmds.mark_for_deployment import mark_for_deployment
from paasta_tools.cli.utils import extract_tags
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import validate_full_git_sha
from paasta_tools.cli.utils import validate_given_deploy_groups
from paasta_tools.deployment_utils import get_currently_deployed_version
from paasta_tools.remote_git import list_remote_refs
from paasta_tools.utils import _log_audit
from paasta_tools.utils import datetime_from_utc_to_local
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeploymentVersion
from paasta_tools.utils import format_table
from paasta_tools.utils import get_git_url
from paasta_tools.utils import list_services
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import parse_timestamp
from paasta_tools.utils import RollbackTypes


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    list_parser = subparsers.add_parser(
        "rollback",
        help="Rollback a docker image to a previous deploy",
        description=(
            "'paasta rollback' is a human-friendly tool for marking a particular "
            "docker image for deployment, which invokes a bounce. While the command "
            "is called 'rollback', it can be used to roll forward or back, as long "
            "as there is a docker image available for the input Git SHA."
        ),
        epilog=(
            "This rollback command uses the Git control plane, which requires network "
            "connectivity as well as authorization to the Git repo.\n\n"
            + PaastaColors.yellow(
                "WARNING: You MUST manually revert changes in Git and go through the normal push process after using this command.\n"
            )
            + PaastaColors.yellow(
                "WARNING: Failing to do so means that Jenkins will redeploy the latest code on the next scheduled build!"
            )
        ),
        # we manually format the epilog to add newlines + give it an attention-grabbing color
        # re: reverting changes in Git post-rollback
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    arg_commit = list_parser.add_argument(
        "-k",
        "--commit",
        help="Git SHA to mark for rollback. "
        "A commit to rollback to is required for paasta rollback to run. However if one is not provided, "
        "paasta rollback will instead output a list of valid git shas to rollback to.",
        required=False,
    )
    arg_commit.completer = lazy_choices_completer(list_previously_deployed_shas)  # type: ignore
    arg_version = list_parser.add_argument(
        "-i",
        "--image-version",
        help="Extra version metadata to mark for rollback. "
        "If your service has enabled no-commit redeploys, both a commit and the extra metadata is required for paasta rollback to run. However if one is not provided, "
        "paasta rollback will instead output a list of valid versions to rollback to.",
        required=False,
        default=None,
    )
    arg_version.completer = lazy_choices_completer(list_previously_deployed_image_versions)  # type: ignore
    arg_deploy_group = list_parser.add_argument(
        "-l",
        "--deploy-groups",
        help="Mark one or more deploy groups to roll back (e.g. "
        '"all.main", "all.main,all.canary"). If no deploy groups specified,'
        "no deploy groups for that service are rolled back. To rollback all deploy groups "
        "use the flag -a or --all-deploy-groups",
        default="",
        required=False,
    )
    arg_deploy_group.completer = lazy_choices_completer(list_deploy_groups)  # type: ignore
    list_parser.add_argument(
        "-a",
        "--all-deploy-groups",
        help="Rollback all deploy groups for the service",
        action="store_true",
        required=False,
    )

    arg_service = list_parser.add_argument(
        "-s",
        "--service",
        help='Name of the service to rollback (e.g. "service1")',
        # strip any potential trailing / for folks tab-completing directories
        type=lambda x: x.rstrip("/"),
    )
    arg_service.completer = lazy_choices_completer(list_services)  # type: ignore
    list_parser.add_argument(
        "-y",
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    list_parser.add_argument(
        "-f",
        "--force",
        help=("Do not check if Git SHA was marked for deployment previously."),
        action="store_true",
    )
    list_parser.set_defaults(command=paasta_rollback)


def list_previously_deployed_shas(
    parsed_args: argparse.Namespace, **kwargs: None
) -> Generator[str, None, None]:
    service = parsed_args.service
    soa_dir = parsed_args.soa_dir
    deploy_groups = {
        deploy_group
        for deploy_group in parsed_args.deploy_groups.split(",")
        if deploy_group
    }
    return (
        version.sha
        for version in get_versions_for_service(service, deploy_groups, soa_dir)
    )


def list_previously_deployed_image_versions(
    parsed_args: argparse.Namespace, **kwargs: None
) -> Generator[str, None, None]:
    service = parsed_args.service
    soa_dir = parsed_args.soa_dir
    deploy_groups = {
        deploy_group
        for deploy_group in parsed_args.deploy_groups.split(",")
        if deploy_group
    }
    return (
        version.image_version
        for version in get_versions_for_service(service, deploy_groups, soa_dir)
    )


def get_versions_for_service(
    service: str, deploy_groups: Collection[str], soa_dir: str
) -> Mapping[DeploymentVersion, Tuple[str, str]]:
    """Returns a dictionary of 2-tuples of the form (timestamp, deploy_group) for each version tuple of (deploy sha, image_version)"""
    if service is None:
        return {}
    git_url = get_git_url(service=service, soa_dir=soa_dir)
    all_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    deploy_groups, _ = validate_given_deploy_groups(all_deploy_groups, deploy_groups)
    previously_deployed_versions: Dict[DeploymentVersion, Tuple[str, str]] = {}
    for ref, sha in list_remote_refs(git_url).items():
        regex_match = extract_tags(ref)
        try:
            deploy_group = regex_match["deploy_group"]
            tstamp = regex_match["tstamp"]
            image_version = regex_match["image_version"]
        except KeyError:
            pass
        else:
            # Now we filter and dedup by picking the most recent sha for a deploy group
            # Note that all strings are greater than ''
            if deploy_group in deploy_groups:
                version = DeploymentVersion(sha=sha, image_version=image_version)
                tstamp_so_far = previously_deployed_versions.get(version, ("all", ""))[
                    1
                ]
                if tstamp > tstamp_so_far:
                    previously_deployed_versions[version] = (tstamp, deploy_group)
    return previously_deployed_versions


def list_previous_versions(
    service: str,
    deploy_groups: Collection[str],
    any_given_deploy_groups: bool,
    versions: Mapping[DeploymentVersion, Tuple],
) -> None:
    def format_timestamp(tstamp: str) -> str:
        return naturaltime(datetime_from_utc_to_local(parse_timestamp(tstamp)))

    print("Below is a list of recent commits:")
    # Latest 10 versions sorted by deployment time
    list_of_versions = sorted(versions.items(), key=lambda x: x[1], reverse=True)[:10]
    rows = [("Timestamp -- UTC", "Human time", "deploy_group", "Version")]
    for version, (timestamp, deploy_group) in list_of_versions:
        rows.extend(
            [(timestamp, format_timestamp(timestamp), deploy_group, repr(version))]
        )
    for line in format_table(rows):
        print(line)
    if len(list_of_versions) >= 2:
        version, (timestamp, deploy_group) = list_of_versions[1]
        deploy_groups_arg_line = (
            "-l %s " % ",".join(deploy_groups) if any_given_deploy_groups else ""
        )
        version_arg = (
            f" --image-version {version.image_version}" if version.image_version else ""
        )
        print(
            "\nFor example, to use the second to last version from {} used on {}, run:".format(
                format_timestamp(timestamp), PaastaColors.bold(deploy_group)
            )
        )
        print(
            PaastaColors.bold(
                f"    paasta rollback -s {service} {deploy_groups_arg_line}-k {version.sha}{version_arg}"
            )
        )


def paasta_rollback(args: argparse.Namespace) -> int:
    """Call mark_for_deployment with rollback parameters
    :param args: contains all the arguments passed onto the script: service,
    deploy groups and sha. These arguments will be verified and passed onto
    mark_for_deployment.
    """

    soa_dir = args.soa_dir
    service = figure_out_service_name(args, soa_dir)

    deploy_info = get_deploy_info(service=service, soa_dir=args.soa_dir)
    if not can_user_deploy_service(deploy_info, service):
        return 1

    git_url = get_git_url(service, soa_dir)

    if args.all_deploy_groups:
        given_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    else:
        given_deploy_groups = {
            deploy_group
            for deploy_group in args.deploy_groups.split(",")
            if deploy_group
        }

    all_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    deploy_groups, invalid = validate_given_deploy_groups(
        all_deploy_groups, given_deploy_groups
    )

    if len(invalid) > 0:
        print(
            PaastaColors.yellow(
                "These deploy groups are not valid and will be skipped: %s.\n"
                % (",").join(invalid)
            )
        )

    if len(deploy_groups) == 0 and not args.all_deploy_groups:
        print(
            PaastaColors.red(
                "ERROR: No valid deploy groups specified for %s.\n Use the flag -a to rollback all valid deploy groups for this service"
                % (service)
            )
        )
        return 1

    versions = get_versions_for_service(service, deploy_groups, soa_dir)
    commit = args.commit
    image_version = args.image_version
    new_version = DeploymentVersion(sha=commit, image_version=image_version)
    if not commit:
        print("Please specify a commit to mark for rollback (-k, --commit).")
        list_previous_versions(
            service, deploy_groups, bool(given_deploy_groups), versions
        )
        return 1
    elif new_version not in versions and not args.force:
        print(
            PaastaColors.red(
                f"This version {new_version} has never been deployed before."
            )
        )
        print("Please double check it or use --force to skip this verification.\n")
        list_previous_versions(
            service, deploy_groups, bool(given_deploy_groups), versions
        )
        return 1

    try:
        validate_full_git_sha(args.commit)
    except argparse.ArgumentTypeError as e:
        print(PaastaColors.red(f"Error: {e}"))
        return 1

    # TODO: Add similar check for when image_version is empty and no-commit redeploys is enforced for requested deploy_group

    returncode = 0

    for deploy_group in deploy_groups:
        rolled_back_from = get_currently_deployed_version(service, deploy_group)
        returncode |= mark_for_deployment(
            git_url=git_url,
            service=service,
            deploy_group=deploy_group,
            commit=commit,
            image_version=image_version,
        )

        # we could also gate this by the return code from m-f-d, but we probably care more about someone wanting to
        # rollback than we care about if the underlying machinery was successfully able to complete the request
        if rolled_back_from != new_version:
            audit_action_details = {
                "rolled_back_from": str(rolled_back_from),
                "rolled_back_to": str(new_version),
                "rollback_type": RollbackTypes.USER_INITIATED_ROLLBACK.value,
                "deploy_group": deploy_group,
            }
            _log_audit(
                action="rollback", action_details=audit_action_details, service=service
            )

    if returncode == 0:
        print(
            PaastaColors.yellow(
                f"WARNING: You MUST manually revert changes in Git! Use 'git revert {rolled_back_from.sha}', and go through the normal push process. "
            )
        )
        print(
            PaastaColors.yellow(
                "WARNING: Failing to do so means that Jenkins will redeploy the latest code on the next scheduled build!"
            )
        )

    return returncode



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import socket
from unittest import mock

import pytest

from paasta_tools import generate_services_file
from paasta_tools import utils


MOCK_NAMESPACES = [
    ("foo.main", {"proxy_port": 1024}),
    ("bar.canary", {"proxy_port": 1025}),
]


@pytest.fixture
def mock_namespaces():
    with mock.patch(
        "paasta_tools.generate_services_file.get_all_namespaces",
        autospec=True,
        return_value=MOCK_NAMESPACES,
    ):
        yield


def test_generate_configuration(mock_namespaces):
    expected = {
        "foo.main": {"host": "169.254.255.254", "port": 1024},
        "bar.canary": {"host": "169.254.255.254", "port": 1025},
    }
    assert expected == generate_services_file.generate_configuration()


@mock.patch("paasta_tools.generate_services_file.parse_args", autospec=True)
@mock.patch.object(socket, "getfqdn", return_value="somehost.yelp", autospec=True)
def test_main_yaml(mock_getfqdn, mock_parse_args, tmpdir, mock_namespaces):
    services_file = tmpdir.join("services.yaml")
    fake_args = mock.Mock()
    fake_args.output_format = "yaml"
    fake_args.output_filename = services_file.strpath
    mock_parse_args.return_value = fake_args

    expected_value = (
        "# This file is automatically generated by paasta_tools.\n"
        "# It was automatically generated at $TIME on somehost.yelp.\n"
        "---\n"
        "bar.canary:\n"
        "  host: 169.254.255.254\n"
        "  port: 1025\n"
        "foo.main:\n"
        "  host: 169.254.255.254\n"
        "  port: 1024\n"
    )

    with mock.patch.object(utils, "datetime") as m:
        m.datetime.now().isoformat.return_value = "$TIME"
        generate_services_file.main()
    assert services_file.read() == expected_value

    # If the only difference is the timestamp, the file should not be regenerated.
    with mock.patch.object(utils, "datetime") as m:
        m.datetime.now().isoformat.return_value = "$TIME+1"
        generate_services_file.main()
    assert services_file.read() == expected_value


@mock.patch("paasta_tools.generate_services_file.parse_args", autospec=True)
def test_main_json(mock_parse_args, tmpdir, mock_namespaces):
    services_file = tmpdir.join("services.json")
    fake_args = mock.Mock()
    fake_args.output_format = "json"
    fake_args.output_filename = services_file.strpath
    mock_parse_args.return_value = fake_args

    generate_services_file.main()

    assert (
        services_file.read()
        == """{
  "bar.canary": {
    "host": "169.254.255.254",
    "port": 1025
  },
  "foo.main": {
    "host": "169.254.255.254",
    "port": 1024
  }
}"""
    )



# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# We just want to test that task_processing is available in the virtualenv
from unittest import mock

from slackclient import SlackClient

from paasta_tools.slack import PaastaSlackClient


@mock.patch("slackclient.SlackClient", autospec=True)
def test_slack_client_doesnt_post_with_no_token(mock_SlackClient):
    psc = PaastaSlackClient(token=None)
    assert psc.post(channels=["foo"], message="bar") == []
    assert mock_SlackClient.api_call.call_count == 0


def test_slack_client_posts_to_multiple_channels():
    fake_sc = mock.create_autospec(SlackClient)
    fake_sc.api_call.side_effect = ({"ok": True}, {"ok": False, "error": "blah"})
    with mock.patch(
        "paasta_tools.slack.SlackClient", autospec=True, return_value=fake_sc
    ):
        psc = PaastaSlackClient(token="fake_token")
        assert psc.post(channels=["1", "2"], message="bar") == [
            {"ok": True},
            {"ok": False, "error": "blah"},
        ]
        assert fake_sc.api_call.call_count == 2, fake_sc.call_args



import json
import os

from paasta_tools.cli.utils import pick_random_port


def main():
    print("-------------------------------------------------------")
    print(
        "Please run export PAASTA_SYSTEM_CONFIG_DIR=etc_paasta_playground to continue"
    )
    print(
        "Please set environment variable PAASTA_TEST_CLUSTER to the cluster you want to use."
    )
    print("-------------------------------------------------------")
    user = os.environ["USER"]
    cluster = os.environ.get("PAASTA_TEST_CLUSTER", f"kind-{user}-k8s-test")
    config_path = "etc_paasta_playground"

    # find unused ports
    port = pick_random_port("paasta-dev-api")

    # Generate api endpoints
    api_endpoints = {"api_endpoints": {cluster: f"http://localhost:{port}"}}
    api_endpoints_path = os.path.join(config_path, "api_endpoints.json")
    with open(api_endpoints_path, "w") as f:
        json.dump(api_endpoints, f)

    # export config path
    os.environ["PAASTA_SYSTEM_CONFIG_DIR"] = config_path

    api_single_process = os.environ.get("PAASTA_API_SINGLE_PROCESS")
    if api_single_process is not None and api_single_process.lower() == "true":
        from paasta_tools.api.api import redirect_argv

        with redirect_argv(["-D", "-c", cluster, str(port)]):
            from paasta_tools.api import api

            api.main("dev-mode")
    else:
        os.execl(
            ".tox/py310-linux/bin/python",
            ".tox/py310-linux/bin/python",
            "-m",
            "paasta_tools.api.api",
            *["-D", "-c", cluster, str(port)],
        )


if __name__ == "__main__":
    main()



from unittest import mock

from paasta_tools.eks_tools import load_eks_service_config
from paasta_tools.eks_tools import load_eks_service_config_no_cache


def test_load_eks_service_config_no_cache():
    with mock.patch(
        "service_configuration_lib.read_service_configuration", autospec=True
    ) as mock_read_service_configuration, mock.patch(
        "paasta_tools.eks_tools.load_service_instance_config", autospec=True
    ) as mock_load_service_instance_config, mock.patch(
        "paasta_tools.eks_tools.load_v2_deployments_json", autospec=True
    ) as mock_load_v2_deployments_json, mock.patch(
        "paasta_tools.eks_tools.EksDeploymentConfig", autospec=True
    ) as mock_eks_deploy_config:
        mock_config = {"freq": "108.9"}
        mock_load_service_instance_config.return_value = mock_config
        mock_read_service_configuration.return_value = {}
        ret = load_eks_service_config_no_cache(
            service="kurupt",
            instance="fm",
            cluster="brentford",
            load_deployments=False,
            soa_dir="/nail/blah",
        )
        mock_load_service_instance_config.assert_called_with(
            service="kurupt",
            instance="fm",
            instance_type="eks",
            cluster="brentford",
            soa_dir="/nail/blah",
        )
        mock_eks_deploy_config.assert_called_with(
            service="kurupt",
            instance="fm",
            cluster="brentford",
            config_dict={"freq": "108.9"},
            branch_dict=None,
            soa_dir="/nail/blah",
        )
        assert not mock_load_v2_deployments_json.called
        assert ret == mock_eks_deploy_config.return_value

        mock_eks_deploy_config.reset_mock()
        ret = load_eks_service_config_no_cache(
            service="kurupt",
            instance="fm",
            cluster="brentford",
            load_deployments=True,
            soa_dir="/nail/blah",
        )
        mock_load_v2_deployments_json.assert_called_with(
            service="kurupt", soa_dir="/nail/blah"
        )
        mock_eks_deploy_config.assert_called_with(
            service="kurupt",
            instance="fm",
            cluster="brentford",
            config_dict={"freq": "108.9"},
            branch_dict=mock_load_v2_deployments_json.return_value.get_branch_dict(),
            soa_dir="/nail/blah",
        )
        assert ret == mock_eks_deploy_config.return_value


def test_load_eks_service_config():
    with mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
    ) as mock_load_kubernetes_service_config_no_cache:
        ret = load_eks_service_config(
            service="kurupt",
            instance="fm",
            cluster="brentford",
            load_deployments=True,
            soa_dir="/nail/blah",
        )
        assert ret == mock_load_kubernetes_service_config_no_cache.return_value



# coding: utf-8

"""
    Paasta API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.3.0
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from paasta_tools.paastaapi.api_client import ApiClient, Endpoint
from paasta_tools.paastaapi.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from paasta_tools.paastaapi.model.resource import Resource


class ResourcesApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __resources(
            self,
            **kwargs
        ):
            """Get resources in the cluster  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.resources(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                groupings ([str]): comma separated list of keys to group by. [optional]
                filter ([str]): List of slave filters in format &#39;filter&#x3D;attr_name:value1,value2&amp;filter&#x3D;attr2:value3,value4&#39;. Matches attr_name&#x3D;(value1 OR value2) AND attr2&#x3D;(value3 OR value4). [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Resource
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        self.resources = Endpoint(
            settings={
                'response_type': (Resource,),
                'auth': [],
                'endpoint_path': '/resources/utilization',
                'operation_id': 'resources',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'groupings',
                    'filter',
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'groupings':
                        ([str],),
                    'filter':
                        ([str],),
                },
                'attribute_map': {
                    'groupings': 'groupings',
                    'filter': 'filter',
                },
                'location_map': {
                    'groupings': 'query',
                    'filter': 'query',
                },
                'collection_format_map': {
                    'groupings': 'csv',
                    'filter': 'multi',
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__resources
        )



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import time
from datetime import datetime
from typing import Optional

import pytz
from pytimeparse.timeparse import timeparse

import paasta_tools.paastaapi.models as paastamodels
from paasta_tools.api import client
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import get_instance_configs_for_service
from paasta_tools.cli.utils import get_paasta_oapi_api_clustername
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_instances
from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.utils import _log_audit
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import PaastaColors


log = logging.getLogger(__name__)


def add_subparser(subparsers):
    autoscale_parser = subparsers.add_parser(
        "autoscale",
        help="Manually scale a service up and down manually, bypassing the normal autoscaler",
    )

    autoscale_parser.add_argument(
        "-s",
        "--service",
        help="Service that you want to autoscale. Like 'example_service'.",
        # strip any potential trailing / for folks tab-completing directories
        type=lambda x: x.rstrip("/"),
    ).completer = lazy_choices_completer(list_services)
    autoscale_parser.add_argument(
        "-i",
        "--instance",
        help="Instance of the service that you want to autoscale. Like 'main' or 'canary'.",
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    autoscale_parser.add_argument(
        "-c",
        "--cluster",
        help="The PaaSTA cluster that has the service instance you want to autoscale. Like 'pnw-prod'.",
        required=True,
    ).completer = lazy_choices_completer(list_clusters)
    autoscale_parser.add_argument(
        "--set", help="Set the number to scale to. Must be an Int.", type=int
    )

    # Temporary override options
    override_group = autoscale_parser.add_argument_group("Temporary Override Options")
    override_group.add_argument(
        "--set-min",
        help="Set the minimum number of replicas (must be >= 1). Requires --for parameter.",
        type=lambda x: int(x)
        if int(x) >= 1
        else autoscale_parser.error("Minimum instances must be >= 1"),
        default=None,
    )
    override_group.add_argument(
        "--set-max",
        help="Set the maximum number of replicas (must be >= 1). Requires --for parameter.",
        type=lambda x: int(x)
        if int(x) >= 1
        else autoscale_parser.error("Maximum instances must be >= 1"),
        default=None,
    )
    override_group.add_argument(
        "--for",
        dest="duration",
        help="Duration for the temporary override (e.g. '3h', '30m'). Required when using --set-min and/or --set-max.",
        default=None,
    )

    autoscale_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    autoscale_parser.set_defaults(command=paasta_autoscale)


def parse_duration_to_seconds(duration: str) -> Optional[int]:
    """Parse a duration string like '3h' or '30m' into seconds.

    Args:
        duration_str: A string representing a duration (e.g., "3h", "30m", "1d")

    Returns:
        The duration in seconds, or None if parsing failed
    """
    if not duration:
        return None

    seconds = timeparse(duration)
    return seconds


def paasta_autoscale(args):
    log.setLevel(logging.DEBUG)
    service = figure_out_service_name(args)

    if (args.set_min is not None or args.set_max is not None) and not args.duration:
        print(
            PaastaColors.yellow(
                "WARNING: --set-min/--set-max usage requires --for parameter to specify duration - defaulting to 30m"
            )
        )
        args.duration = "30m"

    if args.duration is not None and args.set_min is None and args.set_max is None:
        print(
            PaastaColors.red("Error: --for requires --set-min or --set-max parameter")
        )
        return 1

    if args.set is not None and args.set_min is not None and args.set_max is not None:
        print(
            PaastaColors.red(
                "Error: Cannot use both --set and --set-min or --set-max at the same time"
            )
        )
        return 1

    instance_config = next(
        get_instance_configs_for_service(
            service=service,
            soa_dir=args.soa_dir,
            clusters=[args.cluster],
            instances=[args.instance],
        ),
        None,
    )
    if not instance_config:
        print(
            "Could not find config files for this service instance in soaconfigs. Maybe you misspelled an argument?"
        )
        return 1

    api = client.get_paasta_oapi_client(
        cluster=get_paasta_oapi_api_clustername(
            cluster=args.cluster,
            is_eks=(instance_config.__class__ == EksDeploymentConfig),
        ),
        http_res=True,
    )
    if not api:
        print("Could not connect to paasta api. Maybe you misspelled the cluster?")
        return 1

    # TODO: we should probably also make sure we set a couple other defaults - currently, it's possible for
    # status/res/etc to be unbound in some code paths
    err_reason = None
    try:
        # get current autoscaler count
        if args.set is None and args.set_min is None and args.set_max is None:
            log.debug("Getting the current autoscaler count...")
            res, status, _ = api.autoscaler.get_autoscaler_count(
                service=service, instance=args.instance, _return_http_data_only=False
            )

        # set desired instances
        elif args.set is not None:
            log.debug(f"Setting desired instances to {args.set}.")
            msg = paastamodels.AutoscalerCountMsg(desired_instances=int(args.set))
            res, status, _ = api.autoscaler.update_autoscaler_count(
                service=service,
                instance=args.instance,
                autoscaler_count_msg=msg,
                _return_http_data_only=False,
            )

            _log_audit(
                action="manual-scale",
                action_details=str(msg),
                service=service,
                instance=args.instance,
                cluster=args.cluster,
            )

        # set lower bound
        elif args.set_min is not None or args.set_max is not None:
            duration_seconds = parse_duration_to_seconds(args.duration)
            if not duration_seconds:
                print(
                    PaastaColors.red(
                        f"Error: Invalid duration format '{args.duration}'. "
                        f"Please use a format like '3h' or '30m'."
                    )
                )
                return 1
            # NOTE: this is explicitly using time.time() since we're doing everything using epoch time
            # for simplicity
            expiration_time = time.time() + duration_seconds

            log.debug(
                f"Sending the following overrides for duration {args.duration}: min_instances: {args.set_min}, max_instances: {args.set_max}."
            )
            msg = paastamodels.AutoscalingOverride(
                min_instances=args.set_min,
                max_instances=args.set_max,
                expire_after=expiration_time,
            )

            res, status, _ = api.autoscaler.set_autoscaling_override(
                service=service,
                instance=args.instance,
                autoscaling_override=msg,
                _return_http_data_only=False,
            )
            _log_audit(
                action="manual-autoscale-override",
                action_details=str(msg),
                service=service,
                instance=args.instance,
                cluster=args.cluster,
            )
    except api.api_error as exc:
        status = exc.status
        err_reason = exc.body

    if not 200 <= status <= 299:
        print(
            PaastaColors.red(
                f"ERROR: '{args.instance}' is not configured to autoscale OR you set impossible {{min, max}}_instances, "
                f"and `paasta autoscale` could not update it. "
                f"If you want to be able to boost this service, please configure autoscaling for the service "
                f"in its config file by setting min and max instances appropriately. Example: \n"
                f"{args.instance}:\n"
                f"     min_instances: 5\n"
                f"     max_instances: 50\n"
                f"{err_reason}"
            )
        )
        return 0

    log.debug(f"Res: {res} Http: {status}")
    if not args.set_min and not args.set_max:
        print(f"Desired instances: {res.desired_instances}")
    else:
        if args.set_min:
            print(
                f"Temporary override set for {args.service}.{args.instance} with minimum instances: {args.set_min}"
            )
        if args.set_max:
            print(
                f"Temporary override set for {args.service}.{args.instance} with maximum instances: {args.set_max}"
            )
        # folks using this might be in different timezones, so let's convert the expiration time to a few common ones
        # to make it extra clear when the override will expire
        epoch_time = datetime.fromtimestamp(res.expire_after)
        eastern_time = epoch_time.astimezone(pytz.timezone("US/Eastern"))
        pacific_time = epoch_time.astimezone(pytz.timezone("US/Pacific"))
        london_time = epoch_time.astimezone(pytz.timezone("Europe/London"))

        time_format = "%Y-%m-%d %H:%M:%S %Z%z"
        print(f"The {args.duration} override will expire at:")
        print(f"Eastern Time: {eastern_time.strftime(time_format)}")
        print(f"Pacific Time: {pacific_time.strftime(time_format)}")
        print(f"London Time:  {london_time.strftime(time_format)}")

    return 0



from unittest import mock

from paasta_tools import puppet_service_tools
from paasta_tools.utils import compose_job_id


def test_get_puppet_services_that_run_here():
    with mock.patch(
        "os.listdir", autospec=True, return_value=["b", "a"]
    ) as listdir_patch, mock.patch(
        "os.path.exists",
        autospec=True,
        side_effect=lambda x: x
        in ("/etc/nerve/puppet_services.d", "/etc/nerve/puppet_services.d/a"),
    ), mock.patch(
        "paasta_tools.puppet_service_tools.open",
        create=True,
        autospec=None,
        side_effect=mock.mock_open(read_data='{"namespaces": ["main"]}'),
    ):
        puppet_services = puppet_service_tools.get_puppet_services_that_run_here()
        assert puppet_services == {"a": ["main"]}
        listdir_patch.assert_called_once_with(puppet_service_tools.PUPPET_SERVICE_DIR)


def test_get_puppet_services_running_here_for_nerve():
    with mock.patch(
        "paasta_tools.puppet_service_tools.get_puppet_services_that_run_here",
        autospec=True,
        side_effect=lambda: {"d": ["main"], "c": ["main", "canary"]},
    ), mock.patch(
        "paasta_tools.puppet_service_tools._namespaced_get_classic_service_information_for_nerve",
        autospec=True,
        side_effect=lambda x, y, _: (compose_job_id(x, y), {}),
    ):
        assert puppet_service_tools.get_puppet_services_running_here_for_nerve(
            "foo"
        ) == [
            ("c.main", {}),
            ("c.canary", {}),
            ("d.main", {}),
        ]



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import traceback

from pyramid.view import view_config

from paasta_tools.api import settings
from paasta_tools.api.views.exception import ApiFailure
from paasta_tools.kubernetes.remote_run import get_max_job_duration_limit
from paasta_tools.kubernetes.remote_run import remote_run_ready
from paasta_tools.kubernetes.remote_run import remote_run_start
from paasta_tools.kubernetes.remote_run import remote_run_stop
from paasta_tools.kubernetes.remote_run import remote_run_token


DEFAULT_MAX_DURATION = 60 * 60  # 1 hour


@view_config(route_name="remote_run.start", request_method="POST", renderer="json")
def view_remote_run_start(request):
    service = request.swagger_data["service"]
    instance = request.swagger_data["instance"]
    user = request.swagger_data["json_body"]["user"]
    interactive = request.swagger_data["json_body"].get("interactive", True)
    recreate = request.swagger_data["json_body"].get("recreate", False)
    is_toolbox = request.swagger_data["json_body"].get("toolbox", False)
    command = request.swagger_data["json_body"].get("command", None)
    max_duration = min(
        request.swagger_data["json_body"].get("max_duration", DEFAULT_MAX_DURATION),
        get_max_job_duration_limit(),
    )
    try:
        return remote_run_start(
            service=service,
            instance=instance,
            cluster=settings.cluster,
            user=user,
            interactive=interactive,
            recreate=recreate,
            max_duration=max_duration,
            is_toolbox=is_toolbox,
            command=command,
        )
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)


@view_config(route_name="remote_run.poll", request_method="GET", renderer="json")
def view_remote_run_poll(request):
    service = request.swagger_data["service"]
    instance = request.swagger_data["instance"]
    job_name = request.swagger_data["job_name"]
    user = request.swagger_data["user"]
    is_toolbox = request.swagger_data.get("toolbox", False)
    try:
        return remote_run_ready(
            service=service,
            instance=instance,
            cluster=settings.cluster,
            job_name=job_name,
            user=user,
            is_toolbox=is_toolbox,
        )
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)


@view_config(route_name="remote_run.stop", request_method="POST", renderer="json")
def view_remote_run_stop(request):
    service = request.swagger_data["service"]
    instance = request.swagger_data["instance"]
    user = request.swagger_data["json_body"]["user"]
    is_toolbox = request.swagger_data["json_body"].get("toolbox", False)
    try:
        return remote_run_stop(
            service=service,
            instance=instance,
            cluster=settings.cluster,
            user=user,
            is_toolbox=is_toolbox,
        )
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)


@view_config(route_name="remote_run.token", request_method="GET", renderer="json")
def view_remote_run_token(request):
    service = request.swagger_data["service"]
    instance = request.swagger_data["instance"]
    user = request.swagger_data["user"]
    try:
        token = remote_run_token(
            service=service,
            instance=instance,
            cluster=settings.cluster,
            user=user,
        )
        return {"token": token}
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)
